from __future__ import division
import numpy as np
import progressbar
import itertools
from future.utils import string_types
import warnings


class GenomicRegion(object):
    """
    Class representing a genomic region.

    .. attribute:: chromosome

        Name of the chromosome this region is located on

    .. attribute:: start

        Start position of the region in base pairs

    .. attribute:: end

        End position of the region in base pairs

    .. attribute:: strand

        Strand this region is on (+1, -1)

    .. attribute:: ix

        Index of the region in the context of all genomic
        regions.

    """

    def __init__(self, start, end, chromosome=None, ix=None):
        """
        Initialize this object.

        :param start: Start position of the region in base pairs
        :param end: End position of the region in base pairs
        :param chromosome: Name of the chromosome this region is located on
        :param ix: Index of the region in the context of all genomic
                   regions.
        """
        self.start = start
        self.end = end
        self.chromosome = chromosome
        self.ix = ix

    @classmethod
    def from_string(cls, region_string):
        """
        Convert a string into a :class:`~GenomicRegion`.

        This is a very useful convenience function to quickly
        define a :class:`~GenomicRegion` object from a descriptor
        string.

        :param region_string: A string of the form
                              <chromosome>[:<start>-<end>[:<strand>]]
                              (with square brackets indicating optional
                              parts of the string). If any optional
                              part of the string is omitted, intuitive
                              defaults will be chosen.
        :return: :class:`~GenomicRegion`
        """
        chromosome = None
        start = None
        end = None

        # strip whitespace
        no_space_region_string = "".join(region_string.split())
        fields = no_space_region_string.split(':')

        if len(fields) > 3:
            raise ValueError("Genomic range string must be of the form <chromosome>[:<start>-<end>:[<strand>]]")

        # there is chromosome information
        if len(fields) > 0:
            chromosome = fields[0]

        # there is range information
        if len(fields) > 1 and fields[1] != '':
            start_end_bp = fields[1].split('-')
            if len(start_end_bp) > 0:
                try:
                    start = int(start_end_bp[0])
                except ValueError:
                    raise ValueError("Start of genomic range must be integer")

            if len(start_end_bp) > 1:
                try:
                    end = int(start_end_bp[1])
                except ValueError:
                    raise ValueError("End of genomic range must be integer")

                if not end > start:
                    raise ValueError("The end coordinate must be bigger than the start.")

        return cls(start=start, end=end, chromosome=chromosome)

    def overlaps(self, region):
        """
        Check if this region overlaps with the specified region.

        :param region: :class:`~GenomicRegion` object or string
        """
        if isinstance(region, str):
            region = GenomicRegion.from_string(region)

        if region.chromosome != self.chromosome:
            return False

        if region.start <= self.end and region.end >= self.start:
            return True
        return False

    def contains(self, region):
        """
        Check if the specified region is completely contained in this region.

        :param region: :class:`~GenomicRegion` object or string
        """
        if isinstance(region, str):
            region = GenomicRegion.from_string(region)

        if region.chromosome != self.chromosome:
            return False

        if region.start >= self.start and region.end <= self.end:
            return True
        return False

    def _equals(self, region):
        if region.chromosome != self.chromosome:
            return False
        if region.start != self.start:
            return False
        if region.end != self.end:
            return False
        return True

    def __eq__(self, other):
        return self._equals(other)

    def __ne__(self, other):
        return not self._equals(other)

    def copy(self):
        """
        Make a copy of this GenomicRegion object.
        """
        return GenomicRegion(chromosome=self.chromosome, start=self.start, end=self.end)


def sub_regions(regions, region):
    """
    Get regions from a list the overlap with another region.

    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    if isinstance(region, string_types):
        region = GenomicRegion.from_string(region)

    sr = []
    start_ix = None
    end_ix = None
    for i, r in enumerate(regions):
        if r.chromosome == region.chromosome and r.start <= region.end and r.end >= region.start:
            if start_ix is None:
                start_ix = i
            end_ix = i
            sr.append(r.copy())
        else:
            if end_ix is not None:
                break

    if start_ix is None or end_ix is None:
        raise ValueError("Region not found in dataset! {}:{}-{}".format(region.chromosome, region.start, region.end))

    return sr, start_ix, end_ix


def sub_matrix_regions(hic_matrix, regions, region):
    """
    Get a square sub Hi-C matrix that overlaps a given region.

    :param hic_matrix: A square numpy array
    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    sr, start_ix, end_ix = sub_regions(regions, region)

    if start_ix is None:
        return np.empty((0, 0)), sr

    return np.copy(hic_matrix[start_ix:end_ix+1, start_ix:end_ix+1]), sr


def sub_data_regions(data, regions, region):
    """
    Get a sub data matrix that overlaps a given region.

    :param data: a numpy array
    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    sr, start_ix, end_ix = sub_regions(regions, region)

    if start_ix is None:
        return np.empty((0, 0)), sr

    if not isinstance(data, np.ndarray):
        data = np.array(data)

    return np.copy(data[:, start_ix:end_ix+1]), sr


def sub_vector_regions(data, regions, region):
    """
    Get a sub-vector that overlaps a given region

    :param data: a numpy vector
    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    sr, start_ix, end_ix = sub_regions(regions, region)

    if start_ix is None:
        return np.empty((0, 0)), sr

    if not isinstance(data, np.ndarray):
        data = np.array(data)

    return np.copy(data[start_ix:end_ix+1]), sr


def load_matrix(file_name, size=None, sep=None, square=True, ix_converter=None):
    try:  # numpy binary format
        m = np.load(file_name)
    except IOError:  # not an .npy file

        # check number of fields in file
        with open(file_name, 'r') as f:
            line = f.readline()
            while line.startswith('#'):
                line = f.readline()
            line = line.rstrip()
            n_fields = len(line.split(sep))

        if n_fields > 3 or not square:  # square matrix format
            m = np.loadtxt(file_name)
        else:
            if size is None:
                raise ValueError("Must provide matrix size when importing from sparse matrix notation "
                                 "(HicPro, etc)")
            m = np.zeros((size, size))
            with open(file_name, 'r') as f:
                for line in f:
                    if line.startswith("#"):
                        continue
                    line = line.rstrip()
                    fields = line.split(sep)
                    if ix_converter is None:
                        source, sink, weight = int(fields[0]), int(fields[1]), float(fields[2])
                    else:
                        source = ix_converter[fields[0]]
                        sink = ix_converter[fields[1]]
                        weight = float(fields[2])
                    m[source, sink] = weight
                    m[sink, source] = weight

    if square and m.shape[0] != m.shape[1]:
        raise ValueError("Matrix dimensions do not match! ({})".format(m.shape))

    return m


def load_regions(file_name, sep=None):
    regions = []
    ix_converter = None
    with open(file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.rstrip()
            fields = line.split(sep)
            if len(fields) > 2:
                chromosome = fields[0]
                start = int(fields[1]) + 1
                end = int(fields[2])
                ix = i
                if len(fields) > 3 and fields[3] != '.':  # HicPro
                    if ix_converter is None:
                        ix_converter = dict()
                    if fields[3] in ix_converter:
                        raise ValueError("name column in region BED must "
                                         "only contain unique values! ({})".format(fields[3]))
                    ix_converter[fields[3]] = ix
                    start += 1
                regions.append(GenomicRegion(chromosome=chromosome, start=start, end=end, ix=ix))
    return regions, ix_converter


def _get_boundary_distances(regions):
    """
    Return the distances of each region to the boundaries of its chromosome.
    """
    n_bins = len(regions)
    # find distances to chromosome boundaries in bins
    boundary_dist = np.zeros(n_bins, dtype=int)
    last_chromosome = None
    last_chromosome_index = 0
    for i, region in enumerate(regions):
        chromosome = region.chromosome
        if last_chromosome is not None and chromosome != last_chromosome:
            chromosome_length = i-last_chromosome_index
            for j in range(chromosome_length):
                boundary_dist[last_chromosome_index+j] = min(j, i-last_chromosome_index-1-j)
            last_chromosome_index = i
        last_chromosome = chromosome
    chromosome_length = n_bins-last_chromosome_index
    for j in range(chromosome_length):
        boundary_dist[last_chromosome_index+j] = min(j, n_bins-last_chromosome_index-1-j)

    return boundary_dist


def directionality_index(hic, regions=None, window_size=2000000):
    """
    Calculate the directionality index as in Dixon et al. (2012).

    :param hic: A Hi-C matrix in the form of a square numpy array
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    :param window_size: A window size in base pairs
    """
    if regions is None:
        for i in range(hic.shape[0]):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    bin_size = regions[0].end-regions[0].start+1
    bin_distance = int(window_size/bin_size)
    if window_size % bin_size > 0:
        bin_distance += 1

    n_bins = len(regions)
    boundary_dist = _get_boundary_distances(regions)

    left_sums = np.zeros(n_bins)
    right_sums = np.zeros(n_bins)
    for source in range(hic.shape[0]):
        for sink in range(source, hic.shape[1]):
            weight = hic[source, sink]

            if source == sink:
                continue
            if sink - source <= bin_distance:
                if boundary_dist[sink] >= sink-source:
                    left_sums[sink] += weight
                if boundary_dist[source] >= sink-source:
                    right_sums[source] += weight

    di = np.zeros(n_bins)
    for i in range(n_bins):
        A = left_sums[i]
        B = right_sums[i]
        E = (A+B)/2
        if E != 0 and B-A != 0:
            di[i] = ((B-A)/abs(B-A)) * ((((A-E)**2)/E) + (((B-E)**2)/E))

    return di


def masked_matrix(matrix, all_zero=False):
    """
    Returns masked version of HicMatrix. By default, all entries in zero-count
    rows and columns are masked.

    :param matrix: A numpy 2D matrix
    :param all_zero: Mask ALL zero-count entries
    :returns: MaskedArray with zero entries masked
    """
    if all_zero:
        return np.ma.MaskedArray(matrix, mask=np.isclose(matrix, 0.))
    col_zero = np.isclose(np.sum(matrix, axis=0), 0.)
    row_zero = np.isclose(np.sum(matrix, axis=1), 0.)
    mask = np.zeros(matrix.shape, dtype=np.bool_)
    mask[:, col_zero] = np.True_
    mask[row_zero, :] = np.True_
    return np.ma.MaskedArray(matrix, mask=mask)


def kth_diag_indices(n, k):
    """
    Return indices of bins k steps away from the diagonal.
    (from http://stackoverflow.com/questions/10925671/numpy-k-th-diagonal-indices)
    """

    rows, cols = np.diag_indices(n)
    if k < 0:
        return rows[:k], cols[-k:]
    elif k > 0:
        return rows[k:], cols[:-k]
    else:
        return rows, cols


def impute_missing_bins(hic_matrix, regions=None, per_chromosome=True, stat=np.ma.mean):
    """
    Impute missing contacts in a Hi-C matrix.

    For inter-chromosomal data uses the mean of all inter-chromosomal contacts,
    for intra-chromosomal data uses the mean of intra-chromosomal counts at the corresponding diagonal.

    :param hic_matrix: A square numpy array
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    :param per_chromosome: Do imputation on a per-chromosome basis (recommended)
    :param stat: The aggregation statistic to be used for imputation, defaults to the mean.
    """
    if regions is None:
        for i in range(hic_matrix.shape[0]):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    chr_bins = dict()
    for i, region in enumerate(regions):
        if region.chromosome not in chr_bins:
            chr_bins[region.chromosome] = [i, i]
        else:
            chr_bins[region.chromosome][1] = i

    n = len(regions)
    if not hasattr(hic_matrix, "mask"):
        hic_matrix = masked_matrix(hic_matrix)

    imputed = hic_matrix.copy()
    if per_chromosome:
        for c_start, c_end in chr_bins.itervalues():
            # Correcting intrachromoc_startmal contacts by mean contact count at each diagonal
            for i in range(c_end - c_start):
                ind = kth_diag_indices(c_end - c_start, -i)
                diag = imputed[c_start:c_end, c_start:c_end][ind]
                diag[diag.mask] = stat(diag)
                imputed[c_start:c_end, c_start:c_end][ind] = diag
            # Correcting interchromoc_startmal contacts by mean of all contact counts between
            # each set of chromoc_startmes
            for other_start, other_end in chr_bins.itervalues():
                # Only correct upper triangle
                if other_start <= c_start:
                    continue
                inter = imputed[c_start:c_end, other_start:other_end]
                inter[inter.mask] = stat(inter)
                imputed[c_start:c_end, other_start:other_end] = inter
    else:
        for i in range(n):
            diag = imputed[kth_diag_indices(n, -i)]
            diag[diag.mask] = stat(diag)
            imputed[kth_diag_indices(n, -i)] = diag
    # Copying upper triangle to lower triangle
    imputed[np.tril_indices(n)] = imputed.T[np.tril_indices(n)]
    return imputed


def _apply_sliding_func(a, window, func=np.ma.mean):
    """
    Apply function on a sliding window over an array, ignoring Numpy NaN values.

    :param a: Numpy array on which function is applied
    :param window: The sliding window is i - window:i + window + 1
                   so total window is twice this parameter.
    :param func: Aggregation function to apply
    """
    out = np.empty(len(a))
    for i in range(len(a)):
        window_start = max(0, i - window)
        window_end = min(len(a), i + window + 1)
        cur_window = np.array(a[window_start:window_end])
        out[i] = func(cur_window[~np.isnan(cur_window)])
    return out


def insulation_index(hic_matrix, regions=None, window_size=500000, relative=False, mask_thresh=.5,
                     aggr_func=np.nanmean, impute_missing=False, normalize=False,
                     normalization_window=None, gradient=False):
    """
    Calculate the insulation index as in Crane et al. (2015)

    :param hic_matrix: A square numpy array
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    :param window_size: A window size in base pairs
    :param relative: Calculate insulation index relative to upstream and downstream contact strength
    :param mask_thresh: Set the threshold for masked value threshold
    :param aggr_func: Aggregation function used for index calculation, defaults to the mean
    :param impute_missing: Impute missing values, defaults to False
    :param normalize: Normalize index at each region to the mean number of contacts across a larger region
                      and log2-transform data
    :param normalization_window: Window (in number of regions) to normalize contacts to (see normalize).
                                 If None, the whole chromosome will be used for normalization.
    :param gradient: Return the first derivative of the insulation index
    """
    if regions is None:
        for i in range(hic_matrix.shape[0]):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    bin_size = regions[0].end-regions[0].start+1
    d = int(window_size/bin_size)
    if window_size % bin_size > 0:
        d += 1

    chr_bins = dict()
    for i, region in enumerate(regions):
        if region.chromosome not in chr_bins:
            chr_bins[region.chromosome] = [i, i]
        else:
            chr_bins[region.chromosome][1] = i

    if impute_missing:
        hic_matrix = impute_missing_bins(hic_matrix, regions=regions)

    is_masked = False
    if hasattr(hic_matrix, 'mask'):
        is_masked = True

    skipped = 0
    last_chromosome = None
    ins_by_chromosome = []
    for i, r in enumerate(regions):
        if r.chromosome != last_chromosome:
            last_chromosome = r.chromosome
            ins_by_chromosome.append(list())

        # too close to chromosome edge
        if (i - chr_bins[r.chromosome][0] < d or
                chr_bins[r.chromosome][1] - i <= d + 1):
            ins_by_chromosome[-1].append(np.nan)
            # ins_matrix[i] = np.nan
            continue

        # masked region
        if is_masked and hic_matrix.mask[i, i]:
            ins_by_chromosome[-1].append(np.nan)
            # ins_matrix[i] = np.nan
            continue

        up_rel_slice = (slice(i - d, i), slice(i - d, i))
        down_rel_slice = (slice(i + 1, i + d + 1), slice(i + 1, i + d + 1))
        ins_slice = (slice(i + 1, i + d + 1), slice(i - d, i))

        if (is_masked and ((relative and np.sum(hic_matrix.mask[up_rel_slice]) > d*d*mask_thresh) or
                (relative and np.sum(hic_matrix.mask[down_rel_slice]) > d*d*mask_thresh) or
                np.sum(hic_matrix.mask[ins_slice]) > d*d*mask_thresh)):
            # If too close to the edge of chromosome or
            # if more than half of the entries in this quadrant are masked (unmappable)
            # exclude it from the analysis
            skipped += 1
            ins_by_chromosome[-1].append(np.nan)
            # ins_matrix[i] = np.nan
            continue

        if not relative:
            if impute_missing and is_masked:
                s = aggr_func(hic_matrix[ins_slice].data)
                ins_by_chromosome[-1].append(s)
                # ins_matrix[i] = s
            else:
                s = aggr_func(hic_matrix[ins_slice])
                ins_by_chromosome[-1].append(s)
                # ins_matrix[i] = s
        else:
            if not impute_missing and not is_masked:
                s = (aggr_func(hic_matrix[ins_slice]) /
                     aggr_func(np.ma.dstack((hic_matrix[up_rel_slice],
                                             hic_matrix[down_rel_slice]))))
                # ins_matrix[i] = s
                ins_by_chromosome[-1].append(s)
            else:
                s = (aggr_func(hic_matrix[ins_slice].data) /
                     aggr_func(np.ma.dstack((hic_matrix[up_rel_slice].data,
                                             hic_matrix[down_rel_slice].data))))
                # ins_matrix[i] = s
                ins_by_chromosome[-1].append(s)

    if normalize:
        for i in range(len(ins_by_chromosome)):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                if normalization_window is not None:
                    ins_by_chromosome[i] = np.ma.log2(ins_by_chromosome[i]/_apply_sliding_func(
                        ins_by_chromosome[i], normalization_window, func=np.nanmean))
                else:
                    ins_by_chromosome[i] = np.ma.log2(ins_by_chromosome[i]/np.nanmean(ins_by_chromosome[i]))
    if gradient:
        for i in range(len(ins_by_chromosome)):
            ins_by_chromosome[i] = np.gradient(ins_by_chromosome[i])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ins_matrix = np.array(list(itertools.chain.from_iterable(ins_by_chromosome)))
    return ins_matrix


def normalised_insulation_index(hic_matrix, regions=None, window_size=500000, normalisation_window=None):
    return insulation_index(hic_matrix, regions=regions, window_size=window_size, normalize=True,
                            normalization_window=normalisation_window)


def data_array(hic_matrix, regions, tad_method=insulation_index,
               window_sizes=None, **kwargs):
    """
    Calculate an index for a range of window sizes.

    :param hic_matrix: A square numpy array
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    :param tad_method: The method used for index calculation
    :param window_sizes: A list of window sizes in base pairs
    :param kwargs: Parameters passed to tad_method
    """
    if window_sizes is None:
        window_sizes = [int(i) for i in np.logspace(4, 6, 100)]

    if regions is None:
        for i in range(hic_matrix.shape[0]):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    try:
        l = len(window_sizes)
    except TypeError:
        l = None

    pb = progressbar.ProgressBar(max_value=l)
    if l is not None:
        pb.start()

    ws = []
    da = []
    for i, window_size in enumerate(window_sizes):
        if l is not None:
            pb.update(i)
        ws.append(window_size)
        da.append(tad_method(hic_matrix, regions, window_size, **kwargs))

    if l is not None:
        pb.finish()

    return np.array(da), ws


def _border_type(i, values):
    """
    Returns border type:

    :param i: index of potential border
    :param values: insulation index values
    :return: if border: 1 or -1, else 0. 1 if derivative at border is >0, -1 otherwise
    """

    if i == 0:
        return 1
    if i == len(values)-1:
        return -1

    if values[i-1] <= 0 <= values[i+1]:
        return 1
    if values[i+1] <= 0 <= values[i-1]:
        return -1
    return 0


def call_tad_borders(ii_results, cutoff=0):
    """
    Calls TAD borders using the first derivative of the insulation index.

    :param ii_results: (raw) insulation index results, numpy vector
    :param cutoff: raw insulation index threshold for "true" TAD boundaries
    """
    tad_borders = []
    g = np.gradient(ii_results)
    for i in range(len(ii_results)):
        border_type = _border_type(i, g)
        if border_type == 1 and ii_results[i] <= cutoff:
            tad_borders.append(i)
    return tad_borders


def call_tads_insulation_index(ii_results, cutoff, regions=None):
    """
    Call TADs from insulation index vector.

    :param ii_results: (raw) insulation index results, numpy vector
    :param cutoff: raw insulation index threshold for "true" TADs
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    """
    if regions is None:
        regions = []
        for i in range(len(ii_results)):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    if len(regions) != len(ii_results):
        raise ValueError("Insulation index results and regions must be the same size!")

    borders = call_tad_borders(ii_results, cutoff=cutoff)

    tad_regions = []
    previous = None
    for border in borders:
        if previous is not None:
            found_max = False
            for j in range(previous, border+1):
                if ii_results[j] >= cutoff:
                    found_max = True

            if found_max:
                first_region = regions[previous]
                second_region = regions[border]
                if first_region.chromosome != second_region.chromosome:
                    continue
                tad_regions.append(GenomicRegion(chromosome=first_region.chromosome,
                                                 start=first_region.start, end=second_region.end))

        previous = border
    return tad_regions


def call_tads_directionality_index(di_results, cutoff, regions=None):
    """
    Call TADs from directionality index vector.

    :param di_results: (raw) directionality index results, numpy vector
    :param cutoff: raw directionality index threshold for "true" biases
    :param regions: A list of :class:`~GenomicRegion`s - if omitted, will create a dummy list
    """
    if regions is None:
        regions = []
        for i in range(len(di_results)):
            regions.append(GenomicRegion(chromosome='', start=i, end=i))

    tad_regions = []
    last_upstream_bias_region = None
    last_downstream_bias_region = None
    for i, value in enumerate(di_results):
        current_region = regions[i]
        if last_upstream_bias_region is not None and current_region.chromosome != last_upstream_bias_region.chromosome:
            tad_regions.append(GenomicRegion(chromosome=last_upstream_bias_region.chromosome,
                                             start=last_upstream_bias_region.start,
                                             end=regions[i-1].end))
            last_upstream_bias_region = None
            last_downstream_bias_region = None

        # look for upstream bias site
        if value >= cutoff:
            # side-by-side TAD
            if last_downstream_bias_region is not None and last_upstream_bias_region is not None:
                tad = GenomicRegion(chromosome=last_downstream_bias_region.chromosome,
                                    start=last_upstream_bias_region.start,
                                    end=last_downstream_bias_region.end)
                tad_regions.append(tad)
                last_upstream_bias_region = None
                last_downstream_bias_region = None

            # beginning of TAD
            if last_upstream_bias_region is None:
                last_upstream_bias_region = current_region

        if value <= -1*cutoff:
            # potential end of TAD
            if last_upstream_bias_region is not None:
                last_downstream_bias_region = current_region

        if abs(value) < cutoff:
            # back to zero
            if last_downstream_bias_region is not None and last_upstream_bias_region is not None:
                tad = GenomicRegion(chromosome=last_downstream_bias_region.chromosome,
                                    start=last_upstream_bias_region.start,
                                    end=last_downstream_bias_region.end)
                tad_regions.append(tad)
                last_upstream_bias_region = None
                last_downstream_bias_region = None
    return tad_regions



from __future__ import division, print_function
from abc import ABCMeta, abstractmethod
import matplotlib as mpl
mpl.use('TkAgg')
from matplotlib.ticker import MaxNLocator, Formatter, Locator
from matplotlib.widgets import Slider, Button
import matplotlib.patches as patches
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from tadtool.tad import GenomicRegion, sub_matrix_regions, sub_data_regions, \
    data_array, insulation_index, sub_vector_regions, sub_regions, \
    call_tads_insulation_index, directionality_index, call_tads_directionality_index, normalised_insulation_index
import math
import copy
import numpy as np
from bisect import bisect_left
from future.utils import string_types

try:
    import Tkinter as tk
    import tkFileDialog as filedialog
except ImportError:
    import tkinter as tk
    from tkinter import filedialog


class BasePlotter(object):

    __metaclass__ = ABCMeta

    def __init__(self, title):
        self._ax = None
        self.cax = None
        self.title = title

    @abstractmethod
    def _plot(self, region=None, **kwargs):
        raise NotImplementedError("Subclasses need to override _plot function")

    @abstractmethod
    def plot(self, region=None, **kwargs):
        raise NotImplementedError("Subclasses need to override plot function")

    @property
    def fig(self):
        return self._ax.figure

    @property
    def ax(self):
        if not self._ax:
            _, self._ax = plt.subplots()
        return self._ax

    @ax.setter
    def ax(self, value):
        self._ax = value


class GenomeCoordFormatter(Formatter):
    """
    Process axis tick labels to give nice representations
    of genomic coordinates
    """
    def __init__(self, chromosome, display_scale=True):
        """
        :param chromosome: :class:`~kaic.data.genomic.GenomicRegion` or string
        :param display_scale: Boolean
                              Display distance scale at bottom right
        """
        if isinstance(chromosome, GenomicRegion):
            self.chromosome = chromosome.chromosome
        else:
            self.chromosome = chromosome
        self.display_scale = display_scale

    def _format_val(self, x, prec_offset=0):
        if x == 0:
            oom_loc = 0
        else:
            oom_loc = int(math.floor(math.log10(abs(x))))
        view_range = self.axis.axes.get_xlim()
        oom_range = int(math.floor(math.log10(abs(view_range[1] - view_range[0]))))
        if oom_loc >= 3:
            return "{:.{prec}f}kb".format(x/1000, prec=max(0, 3 + prec_offset - oom_range))
        return "{:.0f}b".format(x)

    def __call__(self, x, pos=None):
        """
        Return label for tick at coordinate x. Relative position of
        ticks can be specified with pos. First tick gets chromosome name.
        """
        s = self._format_val(x, prec_offset=1)
        if pos == 0 or x == 0:
            return "{}:{}".format(self.chromosome, s)
        return s

    def get_offset(self):
        """
        Return information about the distances between
        tick bars and the size of the view window.
        Is called by matplotlib and displayed in lower right corner
        of plots.
        """
        if not self.display_scale:
            return ""
        view_range = self.axis.axes.get_xlim()
        view_dist = abs(view_range[1] - view_range[0])
        tick_dist = self.locs[2] - self.locs[1]
        minor_tick_dist = tick_dist/5
        minor_tick_dist_str = self._format_val(minor_tick_dist, prec_offset=2)
        tick_dist_str = self._format_val(tick_dist, prec_offset=1)
        view_dist_str = self._format_val(view_dist)
        return "{}|{}|{}".format(minor_tick_dist_str, tick_dist_str, view_dist_str)


class GenomeCoordLocator(MaxNLocator):
    """
    Choose locations of genomic coordinate ticks on the plot axis.
    Behaves like default Matplotlib locator, except that it always
    places a tick at the start and the end of the window.
    """
    def __call__(self):
        vmin, vmax = self.axis.get_view_interval()
        ticks = self.tick_values(vmin, vmax)
        # Make sure that first and last tick are the start
        # and the end of the genomic range plotted. If next
        # ticks are too close, remove them.
        ticks[0] = vmin
        ticks[-1] = vmax
        if ticks[1] - vmin < (vmax - vmin)/(self._nbins*3):
            ticks = np.delete(ticks, 1)
        if vmax - ticks[-2] < (vmax - vmin)/(self._nbins*3):
            ticks = np.delete(ticks, -2)
        return self.raise_if_exceeds(np.array(ticks))


class MinorGenomeCoordLocator(Locator):
    """
    Choose locations of minor tick marks between major
    tick labels. Modification of the Matplotlib AutoMinorLocator,
    except that it uses the distance between 2nd and 3rd major
    mark as reference, instead of 2nd and 3rd.
    """
    def __init__(self, n):
        self.ndivs = n

    def __call__(self):
        majorlocs = self.axis.get_majorticklocs()
        try:
            majorstep = majorlocs[2] - majorlocs[1]
        except IndexError:
            # Need at least two major ticks to find minor tick locations
            # TODO: Figure out a way to still be able to display minor
            # ticks without two major ticks visible. For now, just display
            # no ticks at all.
            majorstep = 0
        if self.ndivs is None:
            if majorstep == 0:
                # TODO: Need a better way to figure out ndivs
                ndivs = 1
            else:
                x = int(np.round(10 ** (np.log10(majorstep) % 1)))
                if x in [1, 5, 10]:
                    ndivs = 5
                else:
                    ndivs = 4
        else:
            ndivs = self.ndivs
        minorstep = majorstep / ndivs
        vmin, vmax = self.axis.get_view_interval()
        if vmin > vmax:
            vmin, vmax = vmax, vmin
        if len(majorlocs) > 0:
            t0 = majorlocs[1]
            tmin = ((vmin - t0) // minorstep + 1) * minorstep
            tmax = ((vmax - t0) // minorstep + 1) * minorstep
            locs = np.arange(tmin, tmax, minorstep) + t0
            cond = np.abs((locs - t0) % majorstep) > minorstep / 10.0
            locs = locs.compress(cond)
        else:
            locs = []
        return self.raise_if_exceeds(np.array(locs))


class BasePlotter1D(BasePlotter):

    __metaclass__ = ABCMeta

    def __init__(self, title):
        BasePlotter.__init__(self, title=title)

    def plot(self, region=None, ax=None, **kwargs):
        if isinstance(region, string_types):
            region = GenomicRegion.from_string(region)
        if ax:
            self.ax = ax
        # set genome tick formatter
        self.ax.xaxis.set_major_formatter(GenomeCoordFormatter(region))
        self.ax.xaxis.set_major_locator(GenomeCoordLocator(nbins=5))
        self.ax.xaxis.set_minor_locator(MinorGenomeCoordLocator(n=5))
        self.ax.set_title(self.title)
        self._plot(region, **kwargs)
        self.ax.set_xlim(region.start, region.end)
        return self.fig, self.ax


def prepare_normalization(norm="lin", vmin=None, vmax=None):
    if isinstance(norm, mpl.colors.Normalize):
        norm.vmin = vmin
        norm.vmax = vmax
        return norm
    if norm == "log":
        return mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
    elif norm == "lin":
        return mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    else:
        raise ValueError("'{}'' not a valid normalization method.".format(norm))


class BasePlotterHic(object):

    __metaclass__ = ABCMeta

    def __init__(self, hic_matrix, regions=None, colormap='RdBu', norm="log",
                 vmin=None, vmax=None, show_colorbar=True, blend_masked=False):
        if regions is None:
            for i in range(hic_matrix.shape[0]):
                regions.append(GenomicRegion(chromosome='', start=i, end=i))
        self.regions = regions
        self.hic_matrix = hic_matrix

        self.colormap = copy.copy(mpl.cm.get_cmap(colormap))
        if blend_masked:
            self.colormap.set_bad(self.colormap(0))
        self._vmin = vmin
        self._vmax = vmax
        self.norm = prepare_normalization(norm=norm, vmin=vmin, vmax=vmax)
        self.colorbar = None
        self.slider = None
        self.show_colorbar = show_colorbar

    def add_colorbar(self, ax=None):
        ax = self.cax if ax is None else ax
        cmap_data = mpl.cm.ScalarMappable(norm=self.norm, cmap=self.colormap)
        cmap_data.set_array([self.vmin, self.vmax])

        self.colorbar = plt.colorbar(cmap_data, cax=ax, orientation="vertical")

    @property
    def vmin(self):
        return self._vmin if self._vmin else np.nanmin(self.hic_matrix)

    @property
    def vmax(self):
        return self._vmax if self._vmax else np.nanmax(self.hic_matrix)


class HicPlot(BasePlotter1D, BasePlotterHic):
    def __init__(self, hic_matrix, regions=None, title='', colormap='viridis', max_dist=None, norm="log",
                 vmin=None, vmax=None, show_colorbar=True, blend_masked=False):
        BasePlotter1D.__init__(self, title=title)
        BasePlotterHic.__init__(self, hic_matrix, regions=regions, colormap=colormap, vmin=vmin, vmax=vmax,
                                show_colorbar=show_colorbar, norm=norm, blend_masked=blend_masked)
        self.max_dist = max_dist
        self.hicmesh = None

    def _plot(self, region=None, cax=None):
        if region is None:
            raise ValueError("Cannot plot triangle plot for whole genome.")

        hm, sr = sub_matrix_regions(self.hic_matrix, self.regions, region)
        hm[np.tril_indices(hm.shape[0])] = np.nan
        # Remove part of matrix further away than max_dist
        if self.max_dist is not None:
            for i in range(hm.shape[0]):
                i_region = sr[i]
                for j in range(hm.shape[1]):
                    j_region = sr[j]

                    if j_region.start-i_region.end > self.max_dist:
                        hm[i, j] = np.nan

        hm_masked = np.ma.MaskedArray(hm, mask=np.isnan(hm))
        # prepare an array of the corner coordinates of the Hic-matrix
        # Distances have to be scaled by sqrt(2), because the diagonals of the bins
        # are sqrt(2)*len(bin_size)
        sqrt2 = math.sqrt(2)
        bin_coords = np.r_[[(x.start - 1) for x in sr], sr[-1].end]/sqrt2
        X, Y = np.meshgrid(bin_coords, bin_coords)
        # rotatate coordinate matrix 45 degrees
        sin45 = math.sin(math.radians(45))
        X_, Y_ = X*sin45 + Y*sin45, X*sin45 - Y*sin45
        # shift x coords to correct start coordinate and center the diagonal directly on the
        # x-axis
        X_ -= X_[1, 0] - (sr[0].start - 1)
        Y_ -= .5*np.min(Y_) + .5*np.max(Y_)
        # create plot
        self.hicmesh = self.ax.pcolormesh(X_, Y_, hm_masked, cmap=self.colormap, norm=self.norm)

        # set limits and aspect ratio
        #self.ax.set_aspect(aspect="equal")
        ylim_max = 0.5*(region.end-region.start)
        if self.max_dist is not None and self.max_dist/2 < ylim_max:
            ylim_max = self.max_dist/2
        self.ax.set_ylim(0, ylim_max)
        # remove y ticks
        self.ax.set_yticks([])
        # hide background patch
        self.ax.patch.set_visible(False)
        if self.show_colorbar:
            self.add_colorbar(cax)

    def set_clim(self, vmin, vmax):
        self.hicmesh.set_clim(vmin=vmin, vmax=vmax)
        if self.colorbar is not None:
            self.colorbar.set_clim(vmin=vmin, vmax=vmax)
            self.colorbar.draw_all()


class DataArrayPlot(BasePlotter1D):
    def __init__(self, data, window_sizes=None, regions=None, title='', midpoint=None,
                 colormap='coolwarm_r', vmax=None, current_window_size=0, log_y=True):
        if regions is None:
            regions = []
            for i in range(data.shape[1]):
                regions.append(GenomicRegion(chromosome='', start=i, end=i))
        self.regions = regions

        BasePlotter1D.__init__(self, title=title)
        self.da = data
        if window_sizes is None:
            window_sizes = []
            try:
                l = len(data)
            except TypeError:
                l = data.shape[0]
            for i in range(l):
                window_sizes.append(i)

        self.window_sizes = window_sizes
        self.colormap = colormap
        self.midpoint = midpoint
        self.mesh = None
        self.vmax = vmax
        self.window_size_line = None
        self.current_window_size = current_window_size
        self.log_y = log_y

    def _plot(self, region=None, cax=None):
        da_sub, regions_sub = sub_data_regions(self.da, self.regions, region)

        da_sub_masked = np.ma.MaskedArray(da_sub, mask=np.isnan(da_sub))
        bin_coords = np.r_[[(x.start - 1) for x in regions_sub], regions_sub[-1].end]
        x, y = np.meshgrid(bin_coords, self.window_sizes)

        self.mesh = self.ax.pcolormesh(x, y, da_sub_masked, cmap=self.colormap, vmax=self.vmax)
        self.colorbar = plt.colorbar(self.mesh, cax=cax, orientation="vertical")
        self.window_size_line = self.ax.axhline(self.current_window_size, color='red')

        if self.log_y:
            self.ax.set_yscale("log")
        self.ax.set_ylim((np.nanmin(self.window_sizes), np.nanmax(self.window_sizes)))

    def set_clim(self, vmin, vmax):
        self.mesh.set_clim(vmin=vmin, vmax=vmax)
        if self.colorbar is not None:
            self.colorbar.set_clim(vmin=vmin, vmax=vmax)
            self.colorbar.draw_all()

    def update(self, window_size):
        self.window_size_line.set_ydata(window_size)


class TADPlot(BasePlotter1D):
    def __init__(self, regions, title='', color='black'):
        BasePlotter1D.__init__(self, title=title)

        self.regions = regions
        self.color = color
        self.current_region = None

    def _plot(self, region=None, cax=None):
        self.current_region = region
        try:
            sr, start_ix, end_ix = sub_regions(self.regions, region)
            trans = self.ax.get_xaxis_transform()
            for r in sr:
                region_patch = patches.Rectangle(
                    (r.start, .2),
                    width=abs(r.end - r.start), height=.6,
                    transform=trans,
                    facecolor=self.color,
                    edgecolor='white',
                    linewidth=2.
                )
                self.ax.add_patch(region_patch)
        except ValueError:
            pass
        self.ax.axis('off')

    def update(self, regions):
        self.regions = regions
        self.ax.cla()
        self.plot(region=self.current_region, ax=self.ax)


class DataLinePlot(BasePlotter1D):
    def __init__(self, data, regions=None, title='', init_row=0, is_symmetric=False):
        BasePlotter1D.__init__(self, title=title)
        if regions is None:
            regions = []
            for i in range(len(data)):
                regions.append(GenomicRegion(chromosome='', start=i, end=i))

        self.init_row = init_row
        self.data = data
        self.sr = None
        self.da_sub = None
        self.regions = regions
        self.current_region = None
        self.line = None
        self.current_ix = init_row
        self.current_cutoff = None
        self.cutoff_line = None
        self.cutoff_line_mirror = None
        self.is_symmetric = is_symmetric

    def _new_region(self, region):
        self.current_region = region
        self.da_sub, self.sr = sub_data_regions(self.data, self.regions, region)

    def _plot(self, region=None, cax=None):
        self._new_region(region)
        bin_coords = [(x.start - 1) for x in self.sr]
        ds = self.da_sub[self.init_row]
        self.line, = self.ax.plot(bin_coords, ds)
        if not self.is_symmetric:
            self.current_cutoff = (self.ax.get_ylim()[1] - self.ax.get_ylim()[0]) / 2 + self.ax.get_ylim()[0]
        else:
            self.current_cutoff = self.ax.get_ylim()[1]/ 2
        self.ax.axhline(0.0, linestyle='dashed', color='grey')
        self.cutoff_line = self.ax.axhline(self.current_cutoff, color='r')
        if self.is_symmetric:
            self.cutoff_line_mirror = self.ax.axhline(-1*self.current_cutoff, color='r')
        self.ax.set_ylim((np.nanmin(ds), np.nanmax(ds)))

    def update(self, ix=None, cutoff=None, region=None, update_canvas=True):
        if region is not None:
            self._new_region(region)

        if ix is not None and ix != self.current_ix:
            ds = self.da_sub[ix]
            self.current_ix = ix
            self.line.set_ydata(ds)
            self.ax.set_ylim((np.nanmin(ds), np.nanmax(ds)))

            if cutoff is None:
                if not self.is_symmetric:
                    self.update(cutoff=(self.ax.get_ylim()[1]-self.ax.get_ylim()[0])/2 + self.ax.get_ylim()[0],
                                update_canvas=False)
                else:
                    self.update(cutoff=self.ax.get_ylim()[1] / 2, update_canvas=False)

            if update_canvas:
                self.fig.canvas.draw()

        if cutoff is not None and cutoff != self.current_cutoff:
            if self.is_symmetric:
                self.current_cutoff = abs(cutoff)
            else:
                self.current_cutoff = cutoff
            self.cutoff_line.set_ydata(self.current_cutoff)
            if self.is_symmetric:
                self.cutoff_line_mirror.set_ydata(-1*self.current_cutoff)

            if update_canvas:
                self.fig.canvas.draw()


class TADtoolPlot(object):
    def __init__(self, hic_matrix, regions=None, data=None, window_sizes=None, norm='lin', max_dist=3000000,
                 max_percentile=99.99, algorithm='insulation', matrix_colormap=None,
                 data_colormap=None, log_data=True):
        self.hic_matrix = hic_matrix
        if regions is None:
            regions = []
            for i in range(hic_matrix.shape[0]):
                regions.append(GenomicRegion(chromosome='', start=i, end=i))
        self.regions = regions
        self.norm = norm
        self.fig = None
        self.max_dist = max_dist
        self.algorithm = algorithm
        self.svmax = None
        self.min_value = np.nanmin(self.hic_matrix[np.nonzero(self.hic_matrix)])
        self.min_value_data = None
        self.hic_plot = None
        self.tad_plot = None
        self.data_plot = None
        self.line_plot = None
        self.sdata = None
        self.data_ax = None
        self.line_ax = None
        self.da = None
        self.ws = None
        self.current_window_size = None
        self.window_size_text = None
        self.tad_cutoff_text = None
        self.max_percentile = max_percentile
        self.tad_regions = None
        self.current_da_ix = None
        self.button_save_tads = None
        self.button_save_vector = None
        self.button_save_matrix = None

        self.log_data = log_data

        if algorithm == 'insulation':
            self.tad_algorithm = insulation_index
            self.tad_calling_algorithm = call_tads_insulation_index
            self.is_symmetric = False
            if matrix_colormap is None:
                self.matrix_colormap = LinearSegmentedColormap.from_list('myreds', ['white', 'red'])
            if data_colormap is None:
                self.data_plot_color = 'plasma'
        elif algorithm == 'ninsulation':
            self.tad_algorithm = normalised_insulation_index
            self.tad_calling_algorithm = call_tads_insulation_index
            self.is_symmetric = True
            if matrix_colormap is None:
                self.matrix_colormap = LinearSegmentedColormap.from_list('myreds', ['white', 'red'])
            if data_colormap is None:
                self.data_plot_color = LinearSegmentedColormap.from_list('myreds', ['blue', 'white', 'red'])
        elif algorithm == 'directionality':
            self.tad_algorithm = directionality_index
            self.tad_calling_algorithm = call_tads_directionality_index
            self.is_symmetric = True
            if matrix_colormap is None:
                self.matrix_colormap = LinearSegmentedColormap.from_list('myreds', ['white', 'red'])
            if data_colormap is None:
                self.data_plot_color = LinearSegmentedColormap.from_list('myreds', ['blue', 'white', 'red'])

        if data is None:
            self.da, self.ws = data_array(hic_matrix=self.hic_matrix, regions=self.regions,
                                          tad_method=self.tad_algorithm, window_sizes=window_sizes)
        else:
            self.da = data
            if window_sizes is None:
                raise ValueError("window_sizes parameter cannot be None when providing data!")
            self.ws = window_sizes

    def vmax_slider_update(self, val):
        self.hic_plot.set_clim(self.min_value, val)

    def data_slider_update(self, val):
        if self.is_symmetric:
            self.data_plot.set_clim(-1*val, val)
        else:
            self.data_plot.set_clim(self.min_value_data, val)

    def on_click_save_tads(self, event):
        tk.Tk().withdraw()  # Close the root window
        save_path = filedialog.asksaveasfilename()
        if save_path is not None:
            with open(save_path, 'w') as o:
                for region in self.tad_regions:
                    o.write("%s\t%d\t%d\n" % (region.chromosome, region.start-1, region.end))

    def on_click_save_vector(self, event):
        tk.Tk().withdraw()  # Close the root window
        save_path = filedialog.asksaveasfilename()
        if save_path is not None:
            da_sub = self.da[self.current_da_ix]
            with open(save_path, 'w') as o:
                for i, region in enumerate(self.regions):
                    o.write("%s\t%d\t%d\t.\t%e\n" % (region.chromosome, region.start-1, region.end, da_sub[i]))

    def on_click_save_matrix(self, event):
        tk.Tk().withdraw()  # Close the root window
        save_path = filedialog.asksaveasfilename()
        if save_path is not None:
            with open(save_path, 'w') as o:
                # write regions
                for i, region in enumerate(self.regions):
                    o.write("%s:%d-%d" % (region.chromosome, region.start-1, region.end))
                    if i < len(self.regions)-1:
                        o.write("\t")
                    else:
                        o.write("\n")

                # write matrix
                n_rows = self.da.shape[0]
                n_cols = self.da.shape[1]
                for i in range(n_rows):
                    window_size = self.ws[i]
                    o.write("%d\t" % window_size)

                    for j in range(n_cols):
                        o.write("%e" % self.da[i, j])
                        if j < n_cols-1:
                            o.write("\t")
                        else:
                            o.write("\n")

    def plot(self, region=None):
        # set up plotting grid
        self.fig = plt.figure(figsize=(10, 10))

        # main plots
        grid_size = (32, 15)
        hic_vmax_slider_ax = plt.subplot2grid(grid_size, (0, 0), colspan=13)
        hic_ax = plt.subplot2grid(grid_size, (1, 0), rowspan=9, colspan=13)
        hp_cax = plt.subplot2grid(grid_size, (1, 14), rowspan=9, colspan=1)
        tad_ax = plt.subplot2grid(grid_size, (10, 0), rowspan=1, colspan=13, sharex=hic_ax)
        line_ax = plt.subplot2grid(grid_size, (12, 0), rowspan=6, colspan=13, sharex=hic_ax)
        line_cax = plt.subplot2grid(grid_size, (12, 13), rowspan=6, colspan=2)
        data_vmax_slider_ax = plt.subplot2grid(grid_size, (19, 0), colspan=13)
        data_ax = plt.subplot2grid(grid_size, (20, 0), rowspan=9, colspan=13, sharex=hic_ax)
        da_cax = plt.subplot2grid(grid_size, (20, 14), rowspan=9, colspan=1)

        # buttons
        save_tads_ax = plt.subplot2grid(grid_size, (31, 0), rowspan=1, colspan=4)
        self.button_save_tads = Button(save_tads_ax, 'Save TADs')
        self.button_save_tads.on_clicked(self.on_click_save_tads)
        save_vector_ax = plt.subplot2grid(grid_size, (31, 5), rowspan=1, colspan=4)
        self.button_save_vector = Button(save_vector_ax, 'Save current values')
        self.button_save_vector.on_clicked(self.on_click_save_vector)
        save_matrix_ax = plt.subplot2grid(grid_size, (31, 10), rowspan=1, colspan=4)
        self.button_save_matrix = Button(save_matrix_ax, 'Save matrix')
        self.button_save_matrix.on_clicked(self.on_click_save_matrix)

        # add subplot content
        max_value = np.nanpercentile(self.hic_matrix, self.max_percentile)
        init_value = .2*max_value

        # HI-C VMAX SLIDER
        self.svmax = Slider(hic_vmax_slider_ax, 'vmax', self.min_value, max_value, valinit=init_value, color='grey')
        self.svmax.on_changed(self.vmax_slider_update)

        # HI-C
        self.hic_plot = HicPlot(self.hic_matrix, self.regions, max_dist=self.max_dist, norm=self.norm,
                                vmax=init_value, vmin=self.min_value, colormap=self.matrix_colormap)
        self.hic_plot.plot(region, ax=hic_ax, cax=hp_cax)

        # generate data array
        self.min_value_data = np.nanmin(self.da[np.nonzero(self.da)])
        max_value_data = np.nanpercentile(self.da, self.max_percentile)
        init_value_data = .5*max_value_data

        # LINE PLOT
        da_ix = int(self.da.shape[0]/2)
        self.current_da_ix = da_ix
        self.line_plot = DataLinePlot(self.da, regions=self.regions, init_row=da_ix, is_symmetric=self.is_symmetric)
        self.line_plot.plot(region, ax=line_ax)
        self.line_ax = line_ax

        # line info
        self.current_window_size = self.ws[da_ix]
        line_cax.text(.1, .8, 'Window size', fontweight='bold')
        self.window_size_text = line_cax.text(.3, .6, str(self.current_window_size))
        line_cax.text(.1, .4, 'TAD cutoff', fontweight='bold')
        self.tad_cutoff_text = line_cax.text(.3, .2, "%.5f" % self.line_plot.current_cutoff)
        line_cax.axis('off')

        # TAD PLOT
        self.tad_regions = self.tad_calling_algorithm(self.da[da_ix], self.line_plot.current_cutoff, self.regions)

        self.tad_plot = TADPlot(self.tad_regions)
        self.tad_plot.plot(region=region, ax=tad_ax)

        # DATA ARRAY
        self.data_plot = DataArrayPlot(self.da, self.ws, self.regions, vmax=init_value_data,
                                       colormap=self.data_plot_color, current_window_size=self.ws[da_ix],
                                       log_y=self.log_data)
        self.data_plot.plot(region, ax=data_ax, cax=da_cax)

        # DATA ARRAY SLIDER
        if self.is_symmetric:
            self.sdata = Slider(data_vmax_slider_ax, 'vmax', 0.0, max_value_data,
                                valinit=init_value_data, color='grey')
        else:
            self.sdata = Slider(data_vmax_slider_ax, 'vmax', self.min_value_data, max_value_data,
                                valinit=init_value_data, color='grey')

        self.sdata.on_changed(self.data_slider_update)
        self.data_slider_update(init_value_data)

        # clean up
        hic_ax.xaxis.set_visible(False)
        line_ax.xaxis.set_visible(False)

        # enable hover
        self.data_ax = data_ax
        cid = self.fig.canvas.mpl_connect('button_press_event', self.on_click)

        return self.fig, (hic_vmax_slider_ax, hic_ax, line_ax, data_ax, hp_cax, da_cax)

    def on_click(self, event):
        if event.inaxes == self.data_ax or event.inaxes == self.line_ax:
            if event.inaxes == self.data_ax:
                ws_ix = bisect_left(self.ws, event.ydata)
                self.current_window_size = self.ws[ws_ix]
                self.current_da_ix = ws_ix

                self.data_plot.update(window_size=self.ws[ws_ix])
                self.line_plot.update(ix=ws_ix, update_canvas=False)
                self.tad_cutoff_text.set_text("%.5f" % self.line_plot.current_cutoff)
                self.window_size_text.set_text(str(self.current_window_size))
            elif event.inaxes == self.line_ax:
                if self.is_symmetric:
                    self.line_plot.update(cutoff=abs(event.ydata), update_canvas=False)
                else:
                    self.line_plot.update(cutoff=abs(event.ydata), update_canvas=False)
                self.tad_cutoff_text.set_text("%.5f" % self.line_plot.current_cutoff)

            # update TADs
            self.tad_regions = self.tad_calling_algorithm(self.da[self.current_da_ix], self.line_plot.current_cutoff,
                                                          self.regions)
            self.tad_plot.update(self.tad_regions)
            self.fig.canvas.draw()



from setuptools import setup, find_packages

setup(
    name='tadtool',
    version='0.74',
    description='Assistant to find cutoffs in TAD calling algorithms.',
    packages=find_packages(exclude=["test"]),
    install_requires=[
        'numpy>=1.9.0',
        'matplotlib',
        'progressbar2',
        'future'
    ],
    author='Vaquerizas lab',
    author_email='kai.kruse@mpi-muenster.mpg.de',
    url='https://github.com/vaquerizaslab/tadtool',
    download_url='https://github.com/vaquerizaslab/tadtool/tarball/0.74',
    keywords=['bioinformatics', 'hi-c', 'genomics', 'tad'],
    classifiers=[],
    scripts=['bin/tadtool']
)
